//------------------------------------------------------------------------------
//                                                  __________
//                                    _            / ________ \ 
//                                   | |          / /    ___ \ \ 
//                                   | |         / /    |__ \ \ \ 
//                                   | |        / /     __ \ \ \ \ 
//           ________     ________   | |____   /_/ __  /  \ \ \ \ \ 
//          / ______ \   / ______ \  |  ____|      \ \ \__/ / / / /
//         | |      | | | |  ____| | | |            \ \____/ / / /
//         | |      | | | | |_____/  | |             \______/ / /
//         | |      | | | |________  | |________         ____/ /
//         |_|      |_|  \_________|  \_________|       |_____/
//
//                                    NEW ELECTRONIC TECHNOLOGY
//
//------------------------------------------------------------------------------
//
// Copyright (C) 2014 NET GmbH All Rights Reserved.
//
// This document, and its contents are the property of NET GmbH.
// NET GmbH retains all rights to: this document, the document contents,
// and the design techniques expressed by the document contents.
// Neither this document, nor any of the information, nor any of the design
// techniques expressed within may be, either in whole or in part: used,
// reproduced, copied, stored, or conveyed to a third party by any means
// without the express written consent of an authorized representative of
// NET GmbH.
//
//------------------------------------------------------------------------------
// Module     : mem_rdit
// Submodules : non
//
// Purpose    : read init termination
//
// Creator    : Jens Wallner
//------------------------------------------------------------------------------
`timescale 1 ns / 1 ps


module mem_rdit (
    // system signals
    clk,
    rst_n,

    // soft reset
    init_i,

    // read request, extern
    rdreq_ext_ack_i,
    rdreq_ext_addr_o,
    rdreq_ext_blen_o,
    rdreq_ext_val_o,

    // read request, intern
    rdreq_int_ack_o,
    rdreq_int_addr_i,
    rdreq_int_blen_i,
    rdreq_int_val_i,

    // read response, extern
    rdres_ext_data_i,
    rdres_ext_evt_i,

    // read response, intern
    rdres_int_data_o,
    rdres_int_evt_o
  );

  //----------------------------------------------------------------------------
  // port declaration
  //----------------------------------------------------------------------------
  //--------------------------------------------------------
  // system signals
  //--------------------------------------------------------
  input           clk;                  // system clock
  input           rst_n;                // asynchron low active reset

  //--------------------------------------------------------
  // soft reset
  //--------------------------------------------------------
  input           init_i;               // synchron high active

  //--------------------------------------------------------
  // read request, extern
  //--------------------------------------------------------
  input           rdreq_ext_ack_i;      // transaction acknowledge
  output [29:2]   rdreq_ext_addr_o;     // write address
  output [5:0]    rdreq_ext_blen_o;     // burst length
  output          rdreq_ext_val_o;      // transaction valid

  //--------------------------------------------------------
  // read request, intern
  //--------------------------------------------------------
  output          rdreq_int_ack_o;      // transaction acknowledge
  input  [29:2]   rdreq_int_addr_i;     // write address
  input  [5:0]    rdreq_int_blen_i;     // burst length
  input           rdreq_int_val_i;      // transaction valid

  //--------------------------------------------------------
  // read response, extern
  //--------------------------------------------------------
  input  [31:0]   rdres_ext_data_i;     // 8 pels
  input           rdres_ext_evt_i;      // transaction event

  //--------------------------------------------------------
  // read response, intern
  //--------------------------------------------------------
  output [31:0]   rdres_int_data_o;     // 8 pels
  output          rdres_int_evt_o;      // transaction event

  //-----------------------------------------------------
  // Signals -- race buffer rdres_ext_s0=buf(rdres_ext) --
  //-----------------------------------------------------
  // FF
  reg    [31:0]   rdres_ext_s0_data;
  reg             rdres_ext_s0_evt;

  //-----------------------------------------------------
  // Signals -- FHS decouble buffer rdreq_ext=FWBW(rdreq) --
  //-----------------------------------------------------
  // FF
  reg    [29:2]   rdreq_ext_fwbwint_addr;
  reg    [5:0]    rdreq_ext_fwbwint_blen;
  reg    [29:2]   rdreq_ext_fwbwbuf_addr;
  reg    [5:0]    rdreq_ext_fwbwbuf_blen;
  reg             rdreq_ext_fwbwint_val;
  reg             rdreq_ext_fwbwint_ack;// buffered ack
  reg             rdreq_ext_fwbwbuf_val;
  // combinatorial
  wire   [29:2]   rdreq_ext_addr_o;
  wire   [5:0]    rdreq_ext_blen_o;
  wire            rdreq_ack;
  wire            rdreq_ext_val_o;

  //-----------------------------------------------------
  // Signals -- race buffer rdres_int=buf(rdres) --
  //-----------------------------------------------------
  // FF
  reg    [31:0]   rdres_int_data_o;
  reg             rdres_int_evt_o;


  //-----------------------------------------------------
  // Signals -- read completion --
  //-----------------------------------------------------
  // FF
  reg             rc_init;
  reg    [6:0]    rc_cnt;               // read on the flight count
  // combinatorial
  reg             rc_cnt0;
  reg             rdreq_val;
  reg             rdreq_int_ack_o;
  wire   [29:2]   rdreq_addr = rdreq_int_addr_i;
  wire   [5:0]    rdreq_blen = rdreq_int_blen_i;
  reg             rdres_evt;
  wire   [31:0]   rdres_data = rdres_ext_s0_data;
  reg    [6:0]    rc_cnt_upd;

  //-----------------------------------------------------
  // Sync Process  -- read completion --
  //-----------------------------------------------------
  always @(posedge clk or negedge rst_n) begin
    if (!rst_n) begin
      rc_init <= 1'd0;
      rc_cnt  <= 7'd0;
    end
    else begin
      if      (init_i)  rc_init <= 1'd1;
      else if (rc_cnt0) rc_init <= 1'd0;

      rc_cnt <= rc_cnt_upd;
    end
  end

  //-----------------------------------------------------
  // Async Process -- read completion --
  //-----------------------------------------------------
  always @(*) begin
    rc_cnt0         = (rc_cnt==7'd0);

    rdreq_val       = !(rc_init | rc_cnt[6]) & rdreq_int_val_i;
    rdreq_int_ack_o = !(rc_init | rc_cnt[6]) & rdreq_ack;

    rdres_evt       = !rc_init & rdres_ext_s0_evt;

    rc_cnt_upd      = rc_cnt
                    + ( (rdreq_val & rdreq_ack) ? {1'd0,rdreq_int_blen_i} : 7'd0 )
                    - { 6'd0, rdres_ext_s0_evt };
  end


  //-----------------------------------------------------
  // Sync Process  -- race buffer rdres_ext_s0=buf(rdres_ext) --
  //      handshake generation
  //-----------------------------------------------------
  always @(posedge clk or negedge rst_n) begin
    if (!rst_n) begin
      rdres_ext_s0_evt <= 1'b0;
    end
    else begin
      rdres_ext_s0_evt <= rdres_ext_evt_i;
    end
  end
  //-----------------------------------------------------
  // Sync Process  -- race buffer rdres_ext_s0=buf(rdres_ext) --
  //      data path
  //-----------------------------------------------------
  always @(posedge clk) begin
    rdres_ext_s0_data <= rdres_ext_data_i;
  end
  //-----------------------------------------------------
  // Sync Process  -- FHS decouble buffer rdreq_ext=FWBW(rdreq) --
  //-----------------------------------------------------
  always @(posedge clk or negedge rst_n) begin
    if (!rst_n) begin
      rdreq_ext_fwbwint_val <= 1'b0;
      rdreq_ext_fwbwint_ack <= 1'b0;
      rdreq_ext_fwbwbuf_val <= 1'b0;
    end
    else begin
      if (rdreq_ack) rdreq_ext_fwbwint_val <= rdreq_val;
      rdreq_ext_fwbwbuf_val <= rdreq_ext_val_o;
      rdreq_ext_fwbwint_ack <= !rdreq_ext_val_o | rdreq_ext_ack_i;
    end
  end
  //-----------------------------------------------------
  // Sync Process  -- FHS decouble buffer rdreq_ext=FWBW(rdreq) --
  //      data path
  //-----------------------------------------------------
  always @(posedge clk) begin
    if (rdreq_ack) begin
      rdreq_ext_fwbwint_addr <= rdreq_addr;
      rdreq_ext_fwbwint_blen <= rdreq_blen;
    end
    rdreq_ext_fwbwbuf_addr <= rdreq_ext_addr_o;
    rdreq_ext_fwbwbuf_blen <= rdreq_ext_blen_o;
  end

  // bypass mux
  assign rdreq_ext_val_o  = rdreq_ext_fwbwint_ack ? rdreq_ext_fwbwint_val : rdreq_ext_fwbwbuf_val;
  assign rdreq_ext_addr_o = rdreq_ext_fwbwint_ack ? rdreq_ext_fwbwint_addr : rdreq_ext_fwbwbuf_addr;
  assign rdreq_ext_blen_o = rdreq_ext_fwbwint_ack ? rdreq_ext_fwbwint_blen : rdreq_ext_fwbwbuf_blen;

  // acknowledge generation FW
  assign  rdreq_ack = !rdreq_ext_fwbwint_val | rdreq_ext_fwbwint_ack;

  //-----------------------------------------------------
  // Sync Process  -- race buffer rdres_int=buf(rdres) --
  //      handshake generation
  //-----------------------------------------------------
  always @(posedge clk or negedge rst_n) begin
    if (!rst_n) begin
      rdres_int_evt_o <= 1'b0;
    end
    else if (init_i) begin
      rdres_int_evt_o <= 1'b0;
    end
    else begin
      rdres_int_evt_o <= rdres_evt;
    end
  end
  //-----------------------------------------------------
  // Sync Process  -- race buffer rdres_int=buf(rdres) --
  //      data path
  //-----------------------------------------------------
  always @(posedge clk) begin
    rdres_int_data_o <= rdres_data;
  end
endmodule
