//------------------------------------------------------------------------------
//                                                  __________
//                                    _            / ________ \ 
//                                   | |          / /    ___ \ \ 
//                                   | |         / /    |__ \ \ \ 
//                                   | |        / /     __ \ \ \ \ 
//           ________     ________   | |____   /_/ __  /  \ \ \ \ \ 
//          / ______ \   / ______ \  |  ____|      \ \ \__/ / / / /
//         | |      | | | |  ____| | | |            \ \____/ / / /
//         | |      | | | | |_____/  | |             \______/ / /
//         | |      | | | |________  | |________         ____/ /
//         |_|      |_|  \_________|  \_________|       |_____/
//
//                                    NEW ELECTRONIC TECHNOLOGY
//
//------------------------------------------------------------------------------
//
// Copyright (C) 2014 NET GmbH All Rights Reserved.
//
// This document, and its contents are the property of NET GmbH.
// NET GmbH retains all rights to: this document, the document contents,
// and the design techniques expressed by the document contents.
// Neither this document, nor any of the information, nor any of the design
// techniques expressed within may be, either in whole or in part: used,
// reproduced, copied, stored, or conveyed to a third party by any means
// without the express written consent of an authorized representative of
// NET GmbH.
//
//------------------------------------------------------------------------------
// Module     : mem_writ
// Submodules : non
//
// Purpose    : write init termination
//
// Creator    : Jens Wallner
//------------------------------------------------------------------------------
`timescale 1 ns / 1 ps


module mem_writ (
    // system signals
    clk,
    rst_n,

    // soft reset
    init_i,

    // write data, extern
    wrdat_ext_ack_i,
    wrdat_ext_data_o,
    wrdat_ext_val_o,

    // write data, intern
    wrdat_int_ack_o,
    wrdat_int_data_i,
    wrdat_int_val_i,

    // write request, extern
    wrreq_ext_ack_i,
    wrreq_ext_addr_o,
    wrreq_ext_blen_o,
    wrreq_ext_val_o,

    // write request, intern
    wrreq_int_ack_o,
    wrreq_int_addr_i,
    wrreq_int_blen_i,
    wrreq_int_val_i
  );

  //----------------------------------------------------------------------------
  // port declaration
  //----------------------------------------------------------------------------
  //--------------------------------------------------------
  // system signals
  //--------------------------------------------------------
  input           clk;                  // system clock
  input           rst_n;                // asynchron low active reset

  //--------------------------------------------------------
  // soft reset
  //--------------------------------------------------------
  input           init_i;               // synchron high active

  //--------------------------------------------------------
  // write data, extern
  //--------------------------------------------------------
  input           wrdat_ext_ack_i;      // transaction acknowledge
  output [31:0]   wrdat_ext_data_o;     // 8 pels
  output          wrdat_ext_val_o;      // transaction valid

  //--------------------------------------------------------
  // write data, intern
  //--------------------------------------------------------
  output          wrdat_int_ack_o;      // transaction acknowledge
  input  [31:0]   wrdat_int_data_i;     // 8 pels
  input           wrdat_int_val_i;      // transaction valid

  //--------------------------------------------------------
  // write request, extern
  //--------------------------------------------------------
  input           wrreq_ext_ack_i;      // transaction acknowledge
  output [29:2]   wrreq_ext_addr_o;     // write address
  output [5:0]    wrreq_ext_blen_o;     // burst length
  output          wrreq_ext_val_o;      // transaction valid

  //--------------------------------------------------------
  // write request, intern
  //--------------------------------------------------------
  output          wrreq_int_ack_o;      // transaction acknowledge
  input  [29:2]   wrreq_int_addr_i;     // write address
  input  [5:0]    wrreq_int_blen_i;     // burst length
  input           wrreq_int_val_i;      // transaction valid

  //synthesis attribute ram_style of wrreq_ext_sfl_mem is distributed
  //-----------------------------------------------------
  // Signals -- 2^4 entry FIFO wrreq_ext=FIFO(wrreq) --
  //-----------------------------------------------------
  // FF
  reg    [4:0]    wrreq_ext_sfl_cnt;    // number of stored elements
  reg    [3:0]    wrreq_ext_sfl_waddr;  // write address
  reg    [3:0]    wrreq_ext_sfl_raddr;  // read  address
  reg             wrreq_ext_val_o;
  reg    [33:0]   wrreq_ext_sfl_mem[15:0];// memory
  reg    [29:2]   wrreq_ext_addr_o;
  reg    [5:0]    wrreq_ext_blen_o;
  // combinatorial
  reg             wrreq_ack;
  reg             wrreq_ext_sfl_inc;    // memory write
  reg             wrreq_ext_sfl_dec;    // memory read

  //-----------------------------------------------------
  // Signals -- FHS decouble buffer wrdat_ext=FWBW(wrdat) --
  //-----------------------------------------------------
  // FF
  reg    [31:0]   wrdat_ext_fwbwint_data;
  reg    [31:0]   wrdat_ext_fwbwbuf_data;
  reg             wrdat_ext_fwbwint_val;
  reg             wrdat_ext_fwbwint_ack;// buffered ack
  reg             wrdat_ext_fwbwbuf_val;
  // combinatorial
  wire   [31:0]   wrdat_ext_data_o;
  wire            wrdat_ack;
  wire            wrdat_ext_val_o;


  //-----------------------------------------------------
  // Signals -- write completion --
  //-----------------------------------------------------
  // FF
  reg             wc_init;
  reg    [6:0]    wc_cnt;               // incomplete write count
  // combinatorial
  reg             wc_cnt0;
  reg             wrreq_val;
  reg             wrreq_int_ack_o;
  reg             wrdat_val;
  reg             wrdat_int_ack_o;
  reg    [6:0]    wc_cnt_dec;
  wire   [29:2]   wrreq_addr = wrreq_int_addr_i;
  wire   [5:0]    wrreq_blen = wrreq_int_blen_i;
  wire   [31:0]   wrdat_data = wrdat_int_data_i;

  //-----------------------------------------------------
  // Sync Process  -- write completion --
  //-----------------------------------------------------
  always @(posedge clk or negedge rst_n) begin
    if (!rst_n) begin
      wc_init <= 1'd0;
      wc_cnt  <= 7'd0;
    end
    else begin
      if      (init_i)  wc_init <= 1'd1;
      else if (wc_cnt0) wc_init <= 1'd0;

      if (wrreq_val & wrreq_ack) begin
        if (wrdat_val & wrdat_ack) begin
          wc_cnt <= wc_cnt_dec + {1'd0, wrreq_blen};
        end
        else begin
          wc_cnt <= wc_cnt + {1'd0, wrreq_blen};
        end
      end
      else if (wrdat_val & wrdat_ack) begin
        wc_cnt <= wc_cnt_dec;
      end

    end
  end

  //-----------------------------------------------------
  // Async Process -- write completion --
  //-----------------------------------------------------
  always @(*) begin
    wc_cnt0         = (wc_cnt == 7'd0);

    wrreq_val       = !(wc_init | wc_cnt[6]) & wrreq_int_val_i;
    wrreq_int_ack_o = !(wc_init | wc_cnt[6]) & wrreq_ack;

    wrdat_val       = wc_init ? !wc_cnt0 : (!wc_cnt0 & wrdat_int_val_i);
    wrdat_int_ack_o = wc_init ? 1'd0     : (!wc_cnt0 & wrdat_ack);

    wc_cnt_dec      = wc_cnt - 7'd1;
  end

  //-----------------------------------------------------
  // Sync Process  -- 2^4 entry FIFO wrreq_ext=FIFO(wrreq) --
  //      output buffer
  //-----------------------------------------------------
  always @(posedge clk or negedge rst_n) begin
    if (!rst_n) begin
      wrreq_ext_sfl_cnt   <= 5'b00000;
      wrreq_ext_sfl_waddr <= 4'h0;
      wrreq_ext_sfl_raddr <= 4'h0;
      wrreq_ext_val_o     <= 1'b0;
    end
    else if (init_i) begin
      wrreq_ext_sfl_cnt   <= 5'b00000;
      wrreq_ext_sfl_waddr <= 4'h0;
      wrreq_ext_sfl_raddr <= 4'h0;
      wrreq_ext_val_o     <= 1'b0;
    end
    else begin
      wrreq_ext_sfl_cnt   <= wrreq_ext_sfl_cnt + {
        {4{!wrreq_ext_sfl_inc & wrreq_ext_sfl_dec}},
        wrreq_ext_sfl_inc ^ wrreq_ext_sfl_dec
      };
      wrreq_ext_sfl_waddr <= wrreq_ext_sfl_waddr + {3'd0, wrreq_ext_sfl_inc};
      wrreq_ext_sfl_raddr <= wrreq_ext_sfl_raddr + {3'd0, wrreq_ext_sfl_dec};
      if      (wrreq_ext_sfl_dec) wrreq_ext_val_o <= 1'd1;
      else if (wrreq_ext_ack_i)   wrreq_ext_val_o <= 1'd0;
    end
  end
  //-----------------------------------------------------
  // Sync Process  -- 2^4 entry FIFO wrreq_ext=FIFO(wrreq) --
  //      memory
  //-----------------------------------------------------
  always @(posedge clk) begin
    if (wrreq_ext_sfl_inc) wrreq_ext_sfl_mem[wrreq_ext_sfl_waddr] <= {
      wrreq_addr,
      wrreq_blen
    };
    if (wrreq_ext_sfl_dec) {
      wrreq_ext_addr_o,
      wrreq_ext_blen_o
    } <= wrreq_ext_sfl_mem[wrreq_ext_sfl_raddr];
  end
  //-----------------------------------------------------
  // Async Process -- 2^4 entry FIFO wrreq_ext=FIFO(wrreq) --
  //       input handshake
  //-----------------------------------------------------
  always @(*) begin
    wrreq_ack         = !wrreq_ext_sfl_cnt[4];
    wrreq_ext_sfl_inc = wrreq_val & !wrreq_ext_sfl_cnt[4];
  end
  //-----------------------------------------------------
  // Async Process -- 2^4 entry FIFO wrreq_ext=FIFO(wrreq) --
  //       output handshake
  //-----------------------------------------------------
  always @(*) begin
    wrreq_ext_sfl_dec = (|wrreq_ext_sfl_cnt) & (!wrreq_ext_val_o | wrreq_ext_ack_i);
  end
  //-----------------------------------------------------
  // Sync Process  -- FHS decouble buffer wrdat_ext=FWBW(wrdat) --
  //      output buffer
  //-----------------------------------------------------
  always @(posedge clk or negedge rst_n) begin
    if (!rst_n) begin
      wrdat_ext_fwbwint_val <= 1'b0;
      wrdat_ext_fwbwint_ack <= 1'b0;
      wrdat_ext_fwbwbuf_val <= 1'b0;
    end
    else begin
      if (wrdat_ack) wrdat_ext_fwbwint_val <= wrdat_val;
      wrdat_ext_fwbwbuf_val <= wrdat_ext_val_o;
      wrdat_ext_fwbwint_ack <= !wrdat_ext_val_o | wrdat_ext_ack_i;
    end
  end
  //-----------------------------------------------------
  // Sync Process  -- FHS decouble buffer wrdat_ext=FWBW(wrdat) --
  //      data path
  //-----------------------------------------------------
  always @(posedge clk) begin
    if (wrdat_ack) begin
      wrdat_ext_fwbwint_data <= wrdat_data;
    end
    wrdat_ext_fwbwbuf_data <= wrdat_ext_data_o;
  end

  // bypass mux
  assign wrdat_ext_val_o  = wrdat_ext_fwbwint_ack ? wrdat_ext_fwbwint_val : wrdat_ext_fwbwbuf_val;
  assign wrdat_ext_data_o = wrdat_ext_fwbwint_ack ? wrdat_ext_fwbwint_data : wrdat_ext_fwbwbuf_data;

  // acknowledge generation FW
  assign  wrdat_ack = !wrdat_ext_fwbwint_val | wrdat_ext_fwbwint_ack;

endmodule
