//------------------------------------------------------------------------------
//                                                  __________
//                                    _            / ________ \ 
//                                   | |          / /    ___ \ \ 
//                                   | |         / /    |__ \ \ \ 
//                                   | |        / /     __ \ \ \ \ 
//           ________     ________   | |____   /_/ __  /  \ \ \ \ \ 
//          / ______ \   / ______ \  |  ____|      \ \ \__/ / / / /
//         | |      | | | |  ____| | | |            \ \____/ / / /
//         | |      | | | | |_____/  | |             \______/ / /
//         | |      | | | |________  | |________         ____/ /
//         |_|      |_|  \_________|  \_________|       |_____/
//
//                                    NEW ELECTRONIC TECHNOLOGY
//
//------------------------------------------------------------------------------
//
// Copyright (C) 2014 NET GmbH All Rights Reserved.
//
// This document, and its contents are the property of NET GmbH.
// NET GmbH retains all rights to: this document, the document contents,
// and the design techniques expressed by the document contents.
// Neither this document, nor any of the information, nor any of the design
// techniques expressed within may be, either in whole or in part: used,
// reproduced, copied, stored, or conveyed to a third party by any means
// without the express written consent of an authorized representative of
// NET GmbH.
//
//------------------------------------------------------------------------------
// Module     : rw2px32opt
// Submodules : non
//
// Purpose    : read / write interface to Xilinx pX memory port
//
//                - optimized rd/wr scheduling
//                - NOTE: ip requires that connected master provides
//                  write data in advance for multible bursts!
//                - NOTE: write commands shall be back to back
//                - NOTE: read commands shall be back to back
//
// Creator    : iv-tec
//------------------------------------------------------------------------------
`timescale 1 ns / 1 ps


module rw2px32opt (
    // system signals
    clk,
    rst_n,

    // Xilinx MCB command port
    px_cmd_addr_o,
    px_cmd_bl_o,
    px_cmd_en_o,
    px_cmd_full_i,
    px_cmd_instr_o,

    // Xilinx MCB read port
    px_rd_data_i,
    px_rd_empty_i,
    px_rd_en_o,

    // Xilinx MCB write port
    px_wr_data_o,
    px_wr_en_o,
    px_wr_full_i,

    // read request
    rdreq_ack_o,
    rdreq_addr_i,
    rdreq_blen_i,
    rdreq_val_i,

    // raed response
    rdres_data_o,
    rdres_evt_o,

    // write data
    wrdat_ack_o,
    wrdat_data_i,
    wrdat_val_i,

    // write request
    wrreq_ack_o,
    wrreq_addr_i,
    wrreq_blen_i,
    wrreq_val_i
  );

  //----------------------------------------------------------------------------
  // port declaration
  //----------------------------------------------------------------------------
  //--------------------------------------------------------
  // system signals
  //--------------------------------------------------------
  input           clk;                  // system clock
  input           rst_n;                // asynchron low active reset

  //--------------------------------------------------------
  // Xilinx MCB command port
  //--------------------------------------------------------
  output [29:0]   px_cmd_addr_o;        // Byte start address for current
                                        // transaction. Addresses must be
                                        // aligned to port size:
                                        // 32-bit ports: Lower two bits must be
                                        //  0s.
  output [5:0]    px_cmd_bl_o;          // Burst length in number of user words
                                        //  for the current transaction. Burst
                                        // length is encoded as 0 to 63,
                                        // representing 1 to 64 user words [for
                                        //  example, b00011 is a burst length 4
                                        //  transaction]. The user word width
                                        // equals the port width [for example,
                                        // a burst length of 3 on a 64-bit port
                                        //  transfers 3 x 64-bit user words =
                                        // 192 bits total].
  output          px_cmd_en_o;          // This active-High signal is the
                                        // write-enable signal for the Command
                                        // FIFO. This signal is covered in more
                                        //  detail in Chapter 4.
  input           px_cmd_full_i;        // This active-High output is the full
                                        // flag for the Command FIFO. It
                                        // indicates the FIFO cannot accept any
                                        //  more commands and blocks writes to
                                        // the Command FIFO.
  output [2:0]    px_cmd_instr_o;       // Command code for the current
                                        // instruction. Bit 0 represents the
                                        // READ/WRITE select, Bit 1 is Auto
                                        // Precharge enable, and Bit 2
                                        // represents Refresh, which always
                                        // takes priority:
                                        // Write: 3b000
                                        // Read: 3b001
                                        // This signal is covered in more
                                        // detail in Chapter 4.

  //--------------------------------------------------------
  // Xilinx MCB read port
  //--------------------------------------------------------
  input  [31:0]   px_rd_data_i;         // Read Data value returning from
                                        // memory. This signal is driven by the
                                        //  output of the Read Data FIFO into
                                        // FPGA logic. PX_SIZE can be 32, 64,
                                        // or 128 bits, depending on the port
                                        // configuration.
  input           px_rd_empty_i;        // This active-High signal is the empty
                                        //  flag for the Read Data FIFO. It
                                        // indicates there is no valid data in
                                        // the FIFO.
  output          px_rd_en_o;           // This active-High signal is the read
                                        // enable for the Read Data FIFO. Read
                                        // Data is clocked out of the FIFO on
                                        // the rising edge of pX_rd_clk when
                                        // pX_rd_en = 1 and pX_rd_empty = 0.

  //--------------------------------------------------------
  // Xilinx MCB write port
  //--------------------------------------------------------
  output [31:0]   px_wr_data_o;         // Write Data value to be loaded into
                                        // Write Data FIFO and sent to memory.
                                        // PX_SIZE can be 32, 64, or 128 bits,
                                        // depending on port configuration.
  output          px_wr_en_o;           // This active-High signal is the write
                                        //  enable for the Write Data FIFO. It
                                        // indicates that the value on
                                        // pX_wr_data is valid for loading into
                                        //  the FIFO. Data is loaded on the
                                        // rising edge of pX_wr_clk when
                                        // pX_wr_en = 1 and pX_wr_full = 0.
  input           px_wr_full_i;         // This active-High signal is the full
                                        // flag for the Write Data FIFO. When
                                        // this signal is high, it prevents
                                        // data from being loade into the FIFO.

  //--------------------------------------------------------
  // read request
  //--------------------------------------------------------
  output          rdreq_ack_o;          // transaction acknowledge
  input  [29:2]   rdreq_addr_i;         // access address
  input  [5:0]    rdreq_blen_i;         // burst length
  input           rdreq_val_i;          // transaction valid

  //--------------------------------------------------------
  // raed response
  //--------------------------------------------------------
  output [31:0]   rdres_data_o;         // access data
  output          rdres_evt_o;          // transaction event

  //--------------------------------------------------------
  // write data
  //--------------------------------------------------------
  output          wrdat_ack_o;          // transaction acknowledge
  input  [31:0]   wrdat_data_i;         // access data
  input           wrdat_val_i;          // transaction valid

  //--------------------------------------------------------
  // write request
  //--------------------------------------------------------
  output          wrreq_ack_o;          // transaction acknowledge
  input  [29:2]   wrreq_addr_i;         // access address
  input  [5:0]    wrreq_blen_i;         // burst length
  input           wrreq_val_i;          // transaction valid

  //-----------------------------------------------------
  // Signals -- race buffer rdres=buf(rdres_s0) --
  //-----------------------------------------------------
  // FF
  reg    [31:0]   rdres_data_o;
  reg             rdres_evt_o;

  //-----------------------------------------------------
  // Signals -- FHS decouble buffer ac_dat_s0=FWBW(wrdat) --
  //-----------------------------------------------------
  // FF
  reg    [31:0]   ac_dat_s0_fwbwint_data;
  reg    [31:0]   ac_dat_s0_fwbwbuf_data;
  reg             ac_dat_s0_fwbwint_val;
  reg             ac_dat_s0_fwbwint_ack;// buffered ack
  reg             ac_dat_s0_fwbwbuf_val;
  // combinatorial
  wire   [31:0]   ac_dat_s0_data;
  wire            wrdat_ack_o;
  wire            ac_dat_s0_val;

  //-----------------------------------------------------
  // Signals -- assign ac_rd=rdreq --
  //-----------------------------------------------------
  // combinatorial
  wire   [29:2]   ac_rd_addr;
  wire   [5:0]    ac_rd_blen;
  wire            ac_rd_val;            // handshake
  wire            rdreq_ack_o;          // handshake

  //-----------------------------------------------------
  // Signals -- FHS 2:1 fix priority arbiter --
  //-----------------------------------------------------
  // combinatorial
  reg             ac_cmd_s0_arb_s;      // select vector
  reg             ac_cmd_s0_val;
  reg             ac_wr_ack;
  reg             ac_rd_ack;
  reg             ac_cmd_s0_rd;
  reg    [29:2]   ac_cmd_s0_addr;
  reg    [5:0]    ac_cmd_s0_blen;

  //-----------------------------------------------------
  // Signals -- FHS decouble buffer ac_cmd_s1=FWBW(ac_cmd_s0) --
  //-----------------------------------------------------
  // FF
  reg             ac_cmd_s1_fwbwint_rd;
  reg    [29:2]   ac_cmd_s1_fwbwint_addr;
  reg    [5:0]    ac_cmd_s1_fwbwint_bl;
  reg             ac_cmd_s1_fwbwbuf_rd;
  reg    [29:2]   ac_cmd_s1_fwbwbuf_addr;
  reg    [5:0]    ac_cmd_s1_fwbwbuf_bl;
  reg             ac_cmd_s1_fwbwint_val;
  reg             ac_cmd_s1_fwbwint_ack;// buffered ack
  reg             ac_cmd_s1_fwbwbuf_val;
  // combinatorial
  wire            ac_cmd_s1_rd;
  wire   [29:2]   ac_cmd_s1_addr;
  wire   [5:0]    ac_cmd_s1_bl;
  wire            ac_cmd_s0_ack;
  wire            ac_cmd_s1_val;


  parameter WAIT_CNT = 7'd32;
  parameter RD_WAIT  = 4;

  parameter IDLE  = 2'd0,
            WRITE = 2'd1,
            READ  = 2'd2;

  //-----------------------------------------------------
  // Signals -- access control --
  //-----------------------------------------------------
  // FF
  reg    [1:0]    ac_state;
  reg    [7:0]    ac_wr_wait;
  reg             ac_wr_cnt_inc;
  reg    [6:0]    ac_wr_cnt;
  reg    [RD_WAIT:0] ac_rd_wait;
  reg    [6:0]    ac_rd_cnt;            // available entries in MCB response buffer
  // combinatorial
  reg    [1:0]    ac_state_next;
  reg             ac_wr_en;             // write enable
  reg             ac_rd_en;             // read enable
  reg             ac_wr_stop;           // stop write
  reg             ac_wr_req;            // write request
  reg             ac_rd_req;            // read request

  wire            ac_wr_rd = 1'd0;
  wire            ac_rd_rd = 1'd1;

  wire            ac_cmd_s1_ack;
  wire            ac_dat_s0_ack;

  wire   [31:0]   rdres_s0_data;
  wire            rdres_s0_evt;

  wire   [5:0]    ac_cmd_s0_bl;

  wire            ac_wr_val;
  wire   [29:2]   ac_wr_addr;
  wire   [5:0]    ac_wr_blen;


  // command
  assign px_cmd_addr_o  = {ac_cmd_s1_addr, 2'd0};
  assign px_cmd_bl_o    = ac_cmd_s1_bl;
  assign px_cmd_en_o    = ac_cmd_s1_val;
  assign ac_cmd_s1_ack  = !px_cmd_full_i;
  assign px_cmd_instr_o = {2'd0, ac_cmd_s1_rd};

  // write data
  assign px_wr_data_o   = ac_dat_s0_data;
  assign px_wr_en_o     = ac_dat_s0_val;
  assign ac_dat_s0_ack  = !px_wr_full_i;

  // read response
  assign rdres_s0_data = px_rd_data_i;
  assign rdres_s0_evt  =!px_rd_empty_i;
  assign px_rd_en_o    = 1'd1;

  assign ac_cmd_s0_bl  = ac_cmd_s0_blen - 6'd1;

  assign ac_wr_val     = ac_wr_en & wrreq_val_i;
  assign wrreq_ack_o   = ac_wr_en & ac_wr_ack;
  assign ac_wr_addr    = wrreq_addr_i;
  assign ac_wr_blen    = wrreq_blen_i;

  //-----------------------------------------------------
  // Sync Process  -- access control --
  //-----------------------------------------------------
  always @(posedge clk or negedge rst_n) begin
    if (!rst_n) begin
      ac_state       <= IDLE;
      ac_wr_wait     <= 8'd0;
      ac_wr_cnt_inc  <= 1'd0;
      ac_wr_cnt      <= 7'd0;
      ac_rd_wait     <= {RD_WAIT+1{1'd0}};
      ac_rd_cnt      <= 7'd64;
    end
    else begin
      ac_state       <= ac_state_next;
      ac_wr_wait     <= (ac_wr_cnt_inc | !(|ac_wr_cnt)) ? 8'd0 : (ac_wr_wait + {7'd0, !ac_wr_wait[7]});
      ac_wr_cnt_inc  <= wrdat_val_i & wrdat_ack_o;
      ac_wr_cnt      <= ac_wr_cnt
                      - ((wrreq_val_i & wrreq_ack_o) ? {1'd0, wrreq_blen_i} : 7'd0)
                      + {6'd0, ac_wr_cnt_inc};

      ac_rd_wait     <= ac_rd_en
                      ? {RD_WAIT+1{1'd0}}
                      : (ac_rd_wait + {{RD_WAIT{1'd0}}, !ac_rd_wait[RD_WAIT]});
      ac_rd_cnt      <= ac_rd_cnt
                      - ((rdreq_val_i & rdreq_ack_o) ? {1'd0, rdreq_blen_i} : 7'd0)
                      + {6'd0, rdres_evt_o};

    end
  end

  //-----------------------------------------------------
  // Async Process -- access control --
  //-----------------------------------------------------
  always @(*) begin
    ac_wr_stop = ({1'd0, wrreq_blen_i} > ac_wr_cnt) & wrreq_val_i;
    ac_wr_req  = ac_wr_wait[7] | !(ac_wr_cnt < WAIT_CNT);

    ac_rd_req  = ({1'd0,rdreq_blen_i} <= ac_rd_cnt) & rdreq_val_i;

    //----------------------
    // default assignment
    //----------------------
    ac_state_next = ac_state;
    ac_rd_en      = 1'd0;
    ac_wr_en      = 1'd0;

    case( ac_state )
      default: begin
        if      (ac_wr_req) ac_state_next = WRITE;
        else if (ac_rd_req) ac_state_next = READ;
      end
      WRITE:begin
        ac_wr_en = !ac_wr_stop;
        if      (!ac_rd_req & !ac_wr_req)   ac_state_next = IDLE;
        else if (ac_wr_stop | !wrreq_val_i) ac_state_next = READ;
      end
      READ:begin
        ac_rd_en = ac_rd_req;
        if      (!ac_rd_req & !ac_wr_req)                         ac_state_next = IDLE;
        else if (!ac_rd_req | (!ac_rd_wait[RD_WAIT] & ac_wr_req)) ac_state_next = WRITE;
      end
    endcase
  end

  //-----------------------------------------------------
  // Sync Process  -- race buffer rdres=buf(rdres_s0) --
  //      handshake generation
  //      response buffer
  //-----------------------------------------------------
  always @(posedge clk or negedge rst_n) begin
    if (!rst_n) begin
      rdres_evt_o <= 1'b0;
    end
    else begin
      rdres_evt_o <= rdres_s0_evt;
    end
  end
  //-----------------------------------------------------
  // Sync Process  -- race buffer rdres=buf(rdres_s0) --
  //      data path
  //-----------------------------------------------------
  always @(posedge clk) begin
    rdres_data_o <= rdres_s0_data;
  end
  //-----------------------------------------------------
  // Sync Process  -- FHS decouble buffer ac_dat_s0=FWBW(wrdat) --
  //      data buffer
  //-----------------------------------------------------
  always @(posedge clk or negedge rst_n) begin
    if (!rst_n) begin
      ac_dat_s0_fwbwint_val <= 1'b0;
      ac_dat_s0_fwbwint_ack <= 1'b0;
      ac_dat_s0_fwbwbuf_val <= 1'b0;
    end
    else begin
      if (wrdat_ack_o) ac_dat_s0_fwbwint_val <= wrdat_val_i;
      ac_dat_s0_fwbwbuf_val <= ac_dat_s0_val;
      ac_dat_s0_fwbwint_ack <= !ac_dat_s0_val | ac_dat_s0_ack;
    end
  end
  //-----------------------------------------------------
  // Sync Process  -- FHS decouble buffer ac_dat_s0=FWBW(wrdat) --
  //      data path
  //-----------------------------------------------------
  always @(posedge clk) begin
    if (wrdat_ack_o) begin
      ac_dat_s0_fwbwint_data <= wrdat_data_i;
    end
    ac_dat_s0_fwbwbuf_data <= ac_dat_s0_data;
  end

  // bypass mux
  assign ac_dat_s0_val  = ac_dat_s0_fwbwint_ack ? ac_dat_s0_fwbwint_val : ac_dat_s0_fwbwbuf_val;
  assign ac_dat_s0_data = ac_dat_s0_fwbwint_ack ? ac_dat_s0_fwbwint_data : ac_dat_s0_fwbwbuf_data;

  // acknowledge generation FW
  assign  wrdat_ack_o = !ac_dat_s0_fwbwint_val | ac_dat_s0_fwbwint_ack;

  //-----------------------------------------------------
  // Assign -- assign ac_rd=rdreq --
  //-----------------------------------------------------
  assign ac_rd_addr  = rdreq_addr_i;
  assign ac_rd_blen  = rdreq_blen_i;
  assign ac_rd_val   = ac_rd_en & rdreq_val_i;
  assign rdreq_ack_o = ac_rd_en & ac_rd_ack;
  //-----------------------------------------------------
  // Async Process -- FHS 2:1 fix priority arbiter --
  //-----------------------------------------------------
  always @(*) begin
    ac_cmd_s0_arb_s = !ac_wr_val;
    ac_cmd_s0_val = ac_wr_val
                  | ac_rd_val;
    ac_wr_ack       = ac_cmd_s0_ack & (ac_cmd_s0_arb_s == 1'd0);
    ac_rd_ack       = ac_cmd_s0_ack & (ac_cmd_s0_arb_s == 1'd1);
  end
  //-----------------------------------------------------
  // Async Process -- FHS 2:1 fix priority arbiter --
  //-----------------------------------------------------
  always @(*) begin
    ac_cmd_s0_rd   = ac_cmd_s0_arb_s
                   ? ac_rd_rd
                   : ac_wr_rd;
    ac_cmd_s0_addr = ac_cmd_s0_arb_s
                   ? ac_rd_addr
                   : ac_wr_addr;
    ac_cmd_s0_blen = ac_cmd_s0_arb_s
                   ? ac_rd_blen
                   : ac_wr_blen;
  end
  //-----------------------------------------------------
  // Sync Process  -- FHS decouble buffer ac_cmd_s1=FWBW(ac_cmd_s0) --
  //      command buffer
  //-----------------------------------------------------
  always @(posedge clk or negedge rst_n) begin
    if (!rst_n) begin
      ac_cmd_s1_fwbwint_val <= 1'b0;
      ac_cmd_s1_fwbwint_ack <= 1'b0;
      ac_cmd_s1_fwbwbuf_val <= 1'b0;
    end
    else begin
      if (ac_cmd_s0_ack) ac_cmd_s1_fwbwint_val <= ac_cmd_s0_val;
      ac_cmd_s1_fwbwbuf_val <= ac_cmd_s1_val;
      ac_cmd_s1_fwbwint_ack <= !ac_cmd_s1_val | ac_cmd_s1_ack;
    end
  end
  //-----------------------------------------------------
  // Sync Process  -- FHS decouble buffer ac_cmd_s1=FWBW(ac_cmd_s0) --
  //      data path
  //-----------------------------------------------------
  always @(posedge clk) begin
    if (ac_cmd_s0_ack) begin
      ac_cmd_s1_fwbwint_rd   <= ac_cmd_s0_rd;
      ac_cmd_s1_fwbwint_addr <= ac_cmd_s0_addr;
      ac_cmd_s1_fwbwint_bl   <= ac_cmd_s0_bl;
    end
    ac_cmd_s1_fwbwbuf_rd   <= ac_cmd_s1_rd;
    ac_cmd_s1_fwbwbuf_addr <= ac_cmd_s1_addr;
    ac_cmd_s1_fwbwbuf_bl   <= ac_cmd_s1_bl;
  end

  // bypass mux
  assign ac_cmd_s1_val  = ac_cmd_s1_fwbwint_ack ? ac_cmd_s1_fwbwint_val : ac_cmd_s1_fwbwbuf_val;
  assign ac_cmd_s1_rd   = ac_cmd_s1_fwbwint_ack ? ac_cmd_s1_fwbwint_rd : ac_cmd_s1_fwbwbuf_rd;
  assign ac_cmd_s1_addr = ac_cmd_s1_fwbwint_ack ? ac_cmd_s1_fwbwint_addr : ac_cmd_s1_fwbwbuf_addr;
  assign ac_cmd_s1_bl   = ac_cmd_s1_fwbwint_ack ? ac_cmd_s1_fwbwint_bl : ac_cmd_s1_fwbwbuf_bl;

  // acknowledge generation FW
  assign  ac_cmd_s0_ack = !ac_cmd_s1_fwbwint_val | ac_cmd_s1_fwbwint_ack;

endmodule
