-------------------------------------------------------------------------------
-- Project    : GigEpro
-- Title      : Simulation Testbench for Custom Module
-------------------------------------------------------------------------------
-- Copyright (c) 2013 NET New Electronic Technology GmbH
-------------------------------------------------------------------------------
-- Revisions  : Date        Version  Author   Description
--              2013-05-20  1.0      Manfred  Created
--              2015-04-16  2.0      a.kramer Changed for GigEpro
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

library std;
use std.textio.all;

library work;
use work.CustomCfgLib.all;
use work.CustomSimLib.all;
use work.CustomFctLib.all;
use work.WrapperPkg.all;

entity CustomTB is
    -- Empty entity declaration for top level simulation test bench
end CustomTB;

architecture Behavioral of CustomTB is

    component CustomModule is
    port (  
        -- Module Infrastructure Signals
        SysClkxCI         : in  std_logic;              -- Custom Module Clock 125 MHz
        SysRstxRI         : in  std_logic;              -- Custom Module Reset
        ImgRstxRI         : in  std_logic;              -- ImageBus Reset
        
        -- ImageBus Receive Port                
        ImgRxPixValxMI    : in  std_logic;              -- ImageBus Receive Pixel Valid
        ImgRxPixRdyxMO    : out std_logic;              -- ImageBus Receive Pixel Ready - ignore if cIMG_HANDSHAKE_EN = False
        ImgRxPixStatxSI   : in  std_logic_vector(cIMG_STAT_WIDTH-1 downto 0);  -- ImageBus Receive Pixel Status
        ImgRxPixDataxDI   : in  std_logic_vector(cIMG_DATA_WIDTH-1 downto 0);  -- ImageBus Receive Pixel Data

        -- ImageBus Transmit Port              
        ImgTxPixValxMO    : out std_logic;              -- ImageBus Transmit Pixel Valid
        ImgTxPixRdyxMI    : in  std_logic;              -- ImageBus Transmit Pixel Ready - ignore if cIMG_HANDSHAKE_EN = False
        ImgTxPixStatxSO   : out std_logic_vector(cIMG_STAT_WIDTH-1 downto 0);  -- ImageBus Transmit Pixel Status
        ImgTxPixDataxDO   : out std_logic_vector(cIMG_DATA_WIDTH-1 downto 0); -- ImageBus Transmit Pixel Data

        -- SystemBus Control Interface                  
        SysReqxMI         : in  std_logic;              -- SystemBus Cycle Request
        SysAckxMO         : out std_logic;              -- SystemBus Cycle Acknowledge
        SysErrxMO         : out std_logic;              -- SystemBus Cycle Error
        SysWrEnxMI        : in  std_logic;              -- SystemBus Write Enable
        SysAddrxAI        : in  std_logic_vector(cSYS_ADDR_WIDTH-1 downto 0);  -- SystemBus Address
        SysWrDataxDI      : in  std_logic_vector(cSYS_DATA_WIDTH-1 downto 0);  -- SystemBus Write Data
        SysRdDataxDO      : out std_logic_vector(cSYS_DATA_WIDTH-1 downto 0);  -- SystemBus Read Data

        -- Custom Module GPIO
        CustIoCtrlxMO     : out std_logic_vector(1 downto 0); -- GPIO Output Control
        CustIoDataOutxDO  : out std_logic_vector(1 downto 0); -- GPIO Output Data
        CustIoDataInxDI   : in  std_logic_vector(0 downto 0); -- GPIO Input Data
    
        -- MCP command
        px_cmd_addr_o     : out std_logic_vector(29 downto 0);
        px_cmd_bl_o       : out std_logic_vector(5 downto 0);
        px_cmd_en_o       : out std_logic;
        px_cmd_full_i     : in std_logic;
        px_cmd_instr_o    : out std_logic_vector(2 downto 0);
        
        -- MCB read
        px_rd_data_i      : in std_logic_vector(31 downto 0);
        px_rd_empty_i     : in std_logic;
        px_rd_en_o        : out std_logic;
        -- MCP write
        
        px_wr_data_o      : out std_logic_vector(31 downto 0);
        px_wr_en_o        : out std_logic;
        px_wr_full_i      : in std_logic);
        end component;

    component SimParaInit is
    port (  
        -- ImageBus Simulation Parameters
        ImgSimParaxTO     : out tIMG_SIM_PARAMETERS);   
    end component;

    component SysBusStimulus is
    port (  
        -- Infrastructure Signals
        SysClkxCI         : in  std_logic;              -- SystemBus Clock 125 MHz
        SysRstxRI         : in  std_logic;              -- SystemBus Reset
        SimParaxTI        : in  tSIMULATION_PARAMETERS; -- System Simulation Parameters
        DebugPortxTI      : in  std_logic_vector(cDBG_PORT_WIDTH-1 downto 0);   -- Debug Port
    
        -- SystemBus Control Interface                      
        SysReqxMO         : out std_logic;              -- SystemBus Cycle Request
        SysAckxMI         : in  std_logic;              -- SystemBus Cycle Acknowledge
        SysErrxMI         : in  std_logic;              -- SystemBus Cycle Error
        SysIntReqxMI      : in  std_logic;              -- SystemBus Interrupt Request
        SysWrEnxMO        : out std_logic;              -- SystemBus Write Enable
        SysAddrxAO        : out std_logic_vector(cSYS_ADDR_WIDTH-1 downto 0);  -- SystemBus Address
        SysWrDataxDO      : out std_logic_vector(cSYS_DATA_WIDTH-1 downto 0);  -- SystemBus Write Data
        SysRdDataxDI      : in  std_logic_vector(cSYS_DATA_WIDTH-1 downto 0);  -- SystemBus Read Data

        -- Simulation Flow Control
        ImgRstxRO         : out std_logic;              -- ImageBus Reset
        StartImgBusxMO    : out boolean;                -- Start ImageBus frame transfers
        StopImgBusxMO     : out boolean;                -- Stop ImageBus frame transfers
        AbortImgBusxMO    : out boolean);               -- Abort ImageBus frame transfers
    end component;

    component ImgBusStimulus is
    port (  
        -- Infrastructure Signals
        ImgClkxCI         : in  std_logic;              -- ImageBus Clock 125 MHz
        ImgRstxRI         : in  std_logic;              -- ImageBus Reset
        ImgSimParaxTI     : in  tIMG_SIM_PARAMETERS;    -- ImageBus Simulation Parameters

        -- Simulation Flow Control
        StartImgBusxMI    : in  boolean;                -- Start ImageBus frame transfers
        StopImgBusxMI     : in  boolean;                -- Stop ImageBus frame transfers
        AbortImgBusxMI    : in  boolean;                -- Abort ImageBus frame transfers
    
        -- ImageBus Receive Port                
        ImgRxPixValxMO    : out std_logic;              -- ImageBus Receive Pixel Valid
        ImgRxPixRdyxMI    : in  std_logic;              -- ImageBus Receive Pixel Ready - ignore if cIMG_HANDSHAKE_EN = False
        ImgRxPixStatxSO   : out std_logic_vector(cIMG_STAT_WIDTH-1 downto 0);  -- ImageBus Receive Pixel Status
        ImgRxPixDataxDO   : out std_logic_vector(cIMG_DATA_WIDTH-1 downto 0)); -- ImageBus Receive Pixel Data
    end component;

    component ImgBusAnalyzer is
    port (  
        -- Infrastructure Signals
        ImgClkxCI         : in  std_logic;              -- ImageBus Clock 125 MHz
        ImgRstxRI         : in  std_logic;              -- ImageBus Reset
        ImgSimParaxTI     : in  tIMG_SIM_PARAMETERS;    -- ImageBus Simulation Parameters

        -- ImageBus Transmit Port              
        ImgTxPixValxMI    : in  std_logic;              -- ImageBus Transmit Pixel Valid
        ImgTxPixRdyxMO    : out std_logic;              -- ImageBus Transmit Pixel Ready - ignore if cIMG_HANDSHAKE_EN = False
        ImgTxPixStatxSI   : in  std_logic_vector(cIMG_STAT_WIDTH-1 downto 0);  -- ImageBus Transmit Pixel Status
        ImgTxPixDataxDI   : in  std_logic_vector(cIMG_DATA_WIDTH-1 downto 0)); -- ImageBus Transmit Pixel Data
    end component;
    
    component mcb_stim_e is
    port (
        px_mem_clk        : in  std_logic;
        
        -- MCB command
        px_cmd_rstn       : in  std_logic;
        px_cmd_clk_i      : in  std_logic;
        px_cmd_addr_i     : in  std_logic_vector(29 downto 0);
        px_cmd_bl_i       : in  std_logic_vector(5 downto 0);
        px_cmd_en_i       : in  std_logic;
        px_cmd_full_o     : out std_logic;
        px_cmd_instr_i    : in  std_logic_vector(2 downto 0);
        
        -- MCB read
        px_rd_clk_i       : in  std_logic;
        px_rd_data_o      : out std_logic_vector(31 downto 0);
        px_rd_empty_o     : out std_logic;
        px_rd_en_i        : in  std_logic;
        
        -- MCB write
        px_wr_clk_i       : in  std_logic;
        px_wr_data_i      : in  std_logic_vector(31 downto 0);
        px_wr_en_i        : in  std_logic;
        px_wr_full_o      : out std_logic);
  end component;

    -----------------------------------------------------------------------------
    -- Signal Declarations
    -----------------------------------------------------------------------------
    signal ImgSimParaxT     : tIMG_SIM_PARAMETERS;          -- ImageBus Simulation Parameters
    signal StartImgBusxM    : boolean   := False;           -- Start ImageBus frame transfers
    signal StopImgBusxM     : boolean   := False;           -- Stop ImageBus frame transfers
    signal AbortImgBusxM    : boolean   := False;           -- Abort ImageBus frame transfers
    signal DebugPortxT      : std_logic_vector(cDBG_PORT_WIDTH-1 downto 0);   -- Event Logger Port

    -- Infrastructure Signals
    signal SysClkxC         : std_logic := '0';             -- SystemBus Clock 125 MHz
    signal MemClkxC         : std_logic := '0';             -- Memory Clock 312 MHz
    signal SysRstxR         : std_logic := '1';             -- SystemBus Reset (powerup)
    signal SysRstnxR        : std_logic := '0';             -- SystemBus NotReset (powerup)
    signal ImgRstxR         : std_logic := '1';             -- ImageBus Reset

    -- SystemBus Control Interface                  
    signal SysReqxM         : std_logic := '0';             -- SystemBus Cycle Request
    signal SysAckxM         : std_logic := '0';             -- SystemBus Cycle Acknowledge
    signal SysErrxM         : std_logic := '0';             -- SystemBus Cycle Error
    signal SysIntReqxM      : std_logic := '0';             -- SystemBus Interrupt Request
    signal SysWrEnxM        : std_logic := '0';             -- SystemBus Write Enable
    signal SysAddrxA        : std_logic_vector(cSYS_ADDR_WIDTH-1 downto 0) := (others => '0');  -- SystemBus Address
    signal SysWrDataxD      : std_logic_vector(cSYS_DATA_WIDTH-1 downto 0) := (others => '0');  -- SystemBus Write Data
    signal SysRdDataxD      : std_logic_vector(cSYS_DATA_WIDTH-1 downto 0) := (others => '0');  -- SystemBus Read Data
    
    -- Custom Module GPIO
    signal CustIoCtrlxM     : std_logic_vector(1 downto 0); -- GPIO Output Control
    signal CustIoDataOutxD  : std_logic_vector(1 downto 0); -- GPIO Output Data
    signal CustIoDataInxD   : std_logic_vector(0 downto 0); -- GPIO Input Data

    -- ImageBus Receive Port            
    signal ImgRxPixValxM    : std_logic := '0';             -- ImageBus Receive Pixel Valid
    signal ImgRxPixRdyxM    : std_logic := '0';             -- ImageBus Receive Pixel Ready - ignore if cIMG_HANDSHAKE_EN = False
    signal ImgRxPixStatxS   : std_logic_vector(cIMG_STAT_WIDTH-1 downto 0) := (others => '0');  -- ImageBus Receive Pixel Status
    signal ImgRxPixDataxD   : std_logic_vector(cIMG_DATA_WIDTH-1 downto 0) := (others => '0');  -- ImageBus Receive Pixel Data

    -- ImageBus Transmit Port          
    signal ImgTxPixValxM    : std_logic := '0';             -- ImageBus Transmit Pixel Valid
    signal ImgTxPixRdyxM    : std_logic := '0';             -- ImageBus Transmit Pixel Ready - ignore if cIMG_HANDSHAKE_EN = False
    signal ImgTxPixStatxS   : std_logic_vector(cIMG_STAT_WIDTH-1 downto 0) := (others => '0');  -- ImageBus Transmit Pixel Status
    signal ImgTxPixDataxD   : std_logic_vector(cIMG_DATA_WIDTH-1 downto 0) := (others => '0');  -- ImageBus Transmit Pixel Data
    signal ImgTxLastLinexM  : std_logic;                       -- For simulation purposes only
    
    -- MCB command
    signal px_cmd_addr_s    : std_logic_vector(29 downto 0); 
    signal px_cmd_bl_s      : std_logic_vector(5 downto 0);  
    signal px_cmd_en_s      : std_logic;                     
    signal px_cmd_full_s    : std_logic;                     
    signal px_cmd_instr_s   : std_logic_vector(2 downto 0);
    
    -- MCB read             
    signal px_rd_data_s     : std_logic_vector(31 downto 0); 
    signal px_rd_empty_s    : std_logic;                     
    signal px_rd_en_s       : std_logic;
    
    -- MCB write            
    signal px_wr_data_s     : std_logic_vector(31 downto 0); 
    signal px_wr_en_s       : std_logic;                     
    signal px_wr_full_s     : std_logic;

begin

    -- Instantiate Custom Module (UUT) 
    CustModInst: CustomModule
    port map (  
        -- Module Infrastructure Signals
        SysClkxCI         => SysClkxC,
        SysRstxRI         => SysRstxR,
        ImgRstxRI         => ImgRstxR,
        
        -- ImageBus Receive Port                
        ImgRxPixValxMI    => ImgRxPixValxM,
        ImgRxPixRdyxMO    => ImgRxPixRdyxM,
        ImgRxPixStatxSI   => ImgRxPixStatxS,
        ImgRxPixDataxDI   => ImgRxPixDataxD,
    
        -- ImageBus Transmit Port              
        ImgTxPixValxMO    => ImgTxPixValxM, 
        ImgTxPixRdyxMI    => ImgTxPixRdyxM, 
        ImgTxPixStatxSO   => ImgTxPixStatxS,
        ImgTxPixDataxDO   => ImgTxPixDataxD,
    
        -- SystemBus Control Interface                  
        SysReqxMI         => SysReqxM,  
        SysAckxMO         => SysAckxM,  
        SysErrxMO         => SysErrxM,  
        SysWrEnxMI        => SysWrEnxM, 
        SysAddrxAI        => SysAddrxA, 
        SysWrDataxDI      => SysWrDataxD,
        SysRdDataxDO      => SysRdDataxD,

        -- Custom Module GPIO
        CustIoCtrlxMO     => CustIoCtrlxM,   
        CustIoDataOutxDO  => CustIoDataOutxD,
        CustIoDataInxDI   => CustIoDataInxD, 
        
        -- MCB command
        px_cmd_addr_o     => px_cmd_addr_s,
        px_cmd_bl_o       => px_cmd_bl_s,
        px_cmd_en_o       => px_cmd_en_s,
        px_cmd_full_i     => px_cmd_full_s,
        px_cmd_instr_o    => px_cmd_instr_s,
       
        -- MCB read       
        px_rd_data_i      => px_rd_data_s,
        px_rd_empty_i     => px_rd_empty_s,
        px_rd_en_o        => px_rd_en_s,
       
        -- MCB write      
        px_wr_data_o      => px_wr_data_s,
        px_wr_en_o        => px_wr_en_s,
        px_wr_full_i      => px_wr_full_s);
        
    -- Simulation Parameter Setup
    SimParaInst: SimParaInit
    port map (  
        -- ImageBus Simulation Parameters
        ImgSimParaxTO     => ImgSimParaxT);

    -- Setup SystemBus Registers for UUT
    SysStimulus: SysBusStimulus
    port map (  
        -- Infrastructure Signals
        SysClkxCI         => SysClkxC,
        SysRstxRI         => SysRstxR,
        SimParaxTI        => ImgSimParaxT.SimPara,
        DebugPortxTI      => DebugPortxT,
    
        -- SystemBus Control Interface                      
        SysReqxMO         => SysReqxM,  
        SysAckxMI         => SysAckxM,  
        SysErrxMI         => SysErrxM,  
        SysIntReqxMI      => SysIntReqxM,
        SysWrEnxMO        => SysWrEnxM, 
        SysAddrxAO        => SysAddrxA, 
        SysWrDataxDO      => SysWrDataxD,
        SysRdDataxDI      => SysRdDataxD,

        -- Simulation Flow Control
        ImgRstxRO         => ImgRstxR,
        StartImgBusxMO    => StartImgBusxM,
        StopImgBusxMO     => StopImgBusxM,
        AbortImgBusxMO    => AbortImgBusxM);

    -- Generate LVPP input data for UUT
    ImgStimulus: ImgBusStimulus
    port map (  
        -- Infrastructure Signals
        ImgClkxCI         => SysClkxC,
        ImgRstxRI         => ImgRstxR,
        ImgSimParaxTI     => ImgSimParaxT,

        -- Simulation Flow Control
        StartImgBusxMI    => StartImgBusxM,
        StopImgBusxMI     => StopImgBusxM,
        AbortImgBusxMI    => AbortImgBusxM,
    
        -- ImageBus Receive Port                
        ImgRxPixValxMO    => ImgRxPixValxM,
        ImgRxPixRdyxMI    => ImgRxPixRdyxM,
        ImgRxPixStatxSO   => ImgRxPixStatxS,
        ImgRxPixDataxDO   => ImgRxPixDataxD);

    -- Analyze LVPP output data from UUT
    ImgAnalyzer: ImgBusAnalyzer
    port map (  
        -- Infrastructure Signals
        ImgClkxCI         => SysClkxC,
        ImgRstxRI         => ImgRstxR,
        ImgSimParaxTI     => ImgSimParaxT,
    
        -- ImageBus Transmit Port              
        ImgTxPixValxMI    => ImgTxPixValxM, 
        ImgTxPixRdyxMO    => ImgTxPixRdyxM, 
        ImgTxPixStatxSI   => ImgTxPixStatxS,
        ImgTxPixDataxDI   => ImgTxPixDataxD);
        
    MCB_STIM: component mcb_stim_e
    port map(
        px_mem_clk        => MemClkxC,
        
        -- MCB command
        px_cmd_rstn       => SysRstnxR,
        px_cmd_clk_i      => SysClkxC,
        px_cmd_addr_i     => px_cmd_addr_s,
        px_cmd_bl_i       => px_cmd_bl_s,
        px_cmd_en_i       => px_cmd_en_s,
        px_cmd_full_o     => px_cmd_full_s,
        px_cmd_instr_i    => px_cmd_instr_s,
        
        -- MCB read
        px_rd_clk_i       => SysClkxC,
        px_rd_data_o      => px_rd_data_s,
        px_rd_empty_o     => px_rd_empty_s,
        px_rd_en_i        => px_rd_en_s,
        
        -- MCB write
        px_wr_clk_i       => SysClkxC,
        px_wr_data_i      => px_wr_data_s,
        px_wr_en_i        => px_wr_en_s,
        px_wr_full_o      => px_wr_full_s);

    -----------------------------------------------------------------------------
    -- System Clock and Reset
    -----------------------------------------------------------------------------
    PcieRstProc : process
        file     LogFile : text;
    begin
        -- Open and close the log file to delete it's contents (if any).
        if (cLOG_FILE_ENABLE = True) then
            file_open(LogFile, cLOG_FILE_NAME, write_mode); 
            file_close(LogFile);
        end if;

        PrintNow ("         GigEpro Custom Module Testbench");
        PrintNow ("Copyright (c) 2014 NET New Electronic Technology GmbH");
        PrintNow (" ");

        PrintLog ("Executing CustomTB", LogExec);
        SysRstxR <= '1';
        SysRstnxR <= '0';
        wait for 1 ns;

        -- Deassert Reset asynchronously to clock.
        wait for ImgSimParaxT.SimPara.RstPeriod;
        SysRstxR <= '0';
        SysRstnxR <= '1';       
        PrintLog ("[PcieRstProc] Reset deasserted", LogFlowCtl);
        wait;
    end process;

    -- Generate Module clock.
    SysClkxC <= not(SysClkxC) after (1 us/(2*cIMGBUS_PIXEL_FREQ));   -- 125 MHz ImageBus clock
    
    -- Generate Memory clock.
    MemClkxC <= not(MemClkxC) after (1 us/(2*cMEM_FREQ));            -- 312 MHz Memory clock

    -- Abort simulation run when the maximum simulation time has been exceeded.
    SimTimeProc : process
    begin
        wait for 1 us;
        if ((ImgSimParaxT.SimPara.MaxSimTime > 0 ns) and (now >= ImgSimParaxT.SimPara.MaxSimTime)) then
            PrintNow ("Maximum Simulation Time Exceeded (MaxSimTime) - Simulation Terminated");
            report " "
            severity failure;
        end if;
    end process;

    ----------------------------------------------------------------------------
    -- GPIO Ports (not used)
    ----------------------------------------------------------------------------
    CustIoDataInxD <= "0";

    ----------------------------------------------------------------------------
    -- Debug port for simulation flow control
    ----------------------------------------------------------------------------
    -- Detect SOF on the ImageBus input port and EOF on the output port.
    DbgProc: process (SysClkxC, ImgRstxR)
    begin
        if (ImgRstxR = '1') then
            ImgTxLastLinexM <= '0';
        elsif rising_edge(SysClkxC) then
            -- Detect Start-of-Last-Line on the ImageBus Transmit port.
            -- Set LastLine flag and hold until the end of the line.
            if ((ImgTxPixValxM = '1') and (ImgTxPixRdyxM = '1') and (ImgTxPixStatxS = cIMG_STAT_SOLL)) then
                ImgTxLastLinexM <= '1';
            elsif ((ImgTxPixValxM = '1') and (ImgTxPixRdyxM = '1') and (ImgTxPixStatxS /= cIMG_STAT_NONE)) then
                ImgTxLastLinexM <= '0';
            end if;
        end if;
    end process;

    -- Debug(0): Start-of-Frame in ImageBus Receive port
    DebugPortxT(0) <= '1' when ((ImgRxPixValxM = '1') and (ImgRxPixRdyxM = '1') and ((ImgRxPixStatxS = cIMG_STAT_SOF) or (ImgRxPixStatxS = cIMG_STAT_SOSF))) else '0';

    -- Debug(1): End-of-Frame in ImageBus Transmit port
    DebugPortxT(1) <= '1' when ((ImgTxPixValxM = '1') and (ImgTxPixStatxS = cIMG_STAT_EOL) and (ImgTxLastLinexM = '1')) else '0';

    -- Debug(11..2): ImageBus Receive data
    DebugPortxT(11 downto 2) <= ImgRxPixDataxD(11 downto 2) when ((ImgRxPixValxM = '1') and (ImgRxPixRdyxM = '1')) else "0000000000";
    DebugPortxT(cDBG_PORT_WIDTH-1 downto 12) <= (others => '0');

end;
