-- ------------------------------------------------------------------------------
-- File Name         :  $RCSfile:  $
-- Author            :  $Author:  $
-- Revision          :  $Revision:  $
-- Last Modification :  $Date:  $
-- ------------------------------------------------------------------------------
--                                                 __________
--                                   _            / ________ \ 
--                                  | |          / /    ___ \ \ 
--                                  | |         / /    |__ \ \ \ 
--                                  | |        / /     __ \ \ \ \ 
--          ________     ________   | |____   /_/ __  /  \ \ \ \ \ 
--         / ______ \   / ______ \  |  ____|      \ \ \__/ / / / /
--        | |      | | | |  ____| | | |            \ \____/ / / /
--        | |      | | | | |_____/  | |             \______/ / /
--        | |      | | | |________  | |________         ____/ /
--        |_|      |_|  \_________|  \_________|       |_____/
--
--                                   NEW ELECTRONIC TECHNOLOGY
--
-- ------------------------------------------------------------------------------
--
-- Copyright (C) 2014 NET GmbH All Rights Reserved.
--
-- This document, and its contents are the property of NET GmbH.
-- NET GmbH retains all rights to: this document, the document contents,
-- and the design techniques expressed by the document contents.
-- Neither this document, nor any of the information, nor any of the design
-- techniques expressed within may be, either in whole or in part: used,
-- reproduced, copied, stored, or conveyed to a third party by any means
-- without the express written consent of an authorized representative of
-- NET GmbH.
--
-- ------------------------------------------------------------------------------
-- Module     : mcb_stim
-- Submodules : non
--
-- Purpose    :
--
-- Creator    : a.kramer
-- ------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.numeric_std.all;

library work;
use work.CustomCfgLib.all;

entity mcb_stim_e is
  port(
    px_mem_clk:           in std_logic;
    px_cmd_rstn:          in std_logic;
    -- MCB command
    px_cmd_clk_i:         in std_logic;
    px_cmd_addr_i:        in std_logic_vector(29 downto 0);
    px_cmd_bl_i:          in std_logic_vector(5 downto 0);
    px_cmd_en_i:          in std_logic;
    px_cmd_full_o:        out std_logic;
    px_cmd_instr_i:       in std_logic_vector(2 downto 0);
    -- MCB read
    px_rd_clk_i:          in std_logic;
    px_rd_data_o:         out std_logic_vector(31 downto 0);
    px_rd_empty_o:        out std_logic;
    px_rd_en_i:           in std_logic;
    -- MCB write
    px_wr_clk_i:          in std_logic;
    px_wr_data_i:         in std_logic_vector(31 downto 0);
    px_wr_en_i:           in std_logic;
    px_wr_full_o:         out std_logic
  	);	
end mcb_stim_e;


architecture mcb_stim_a of mcb_stim_e is

  type cmd is record
    addr:       std_logic_vector(27 downto 0);
    bl:         std_logic_vector(5 downto 0);
    instr:      std_logic_vector(2 downto 0);
  end record;
  
  constant cMCB_CMD_BUF_SIZE:     integer := 4;
  constant cMCB_WRRD_BUF_SIZE:    integer := 64;
  
  type cmd_buff is array (0 to cMCB_CMD_BUF_SIZE-1) of cmd;
  type buff is array (0 to cMCB_WRRD_BUF_SIZE-1) of std_logic_vector(31 downto 0);
  type mem is array (0 to cMEM_SIZE-1) of std_logic_vector(31 downto 0);  
  type state_typ is (idle, wr, rd);
  
  signal fsm_state:               state_typ;
  signal fsm_nextstate:           state_typ;
  signal fsm_wr_evt:              std_logic;
  signal fsm_rd_evt:              std_logic;
  signal fsm_addr_offs_rst:       std_logic;
  signal fsm_cmd_finish_evt:      std_logic;
  
  signal cmd_buffer:              cmd_buff := (others => (others => (others => '0')));
  signal cmd_inptr:               integer range 0 to cMCB_CMD_BUF_SIZE-1;
  signal cmd_outptr:              integer range 0 to cMCB_CMD_BUF_SIZE-1;
  
  signal wr_buffer:               buff;
  signal wr_inptr:                integer range 0 to cMCB_WRRD_BUF_SIZE-1;
  signal wr_outptr:               integer range 0 to cMCB_WRRD_BUF_SIZE-1;
  
  signal rd_buffer:               buff;
  signal rd_inptr:                integer range 0 to cMCB_WRRD_BUF_SIZE-1;
  signal rd_outptr:               integer range 0 to cMCB_WRRD_BUF_SIZE-1;
  
  signal mem_data:                mem;
  signal mem_addr:                integer range 0 to cMEM_SIZE-1;
  signal mem_addr_start:          integer range 0 to cMEM_SIZE-1;
  signal mem_addr_offs:           integer range 0 to 100;
  
 
begin

  fsm_comb: process(fsm_state, cmd_outptr, cmd_inptr, mem_addr_offs)
  begin
    fsm_nextstate <= fsm_state;
    fsm_wr_evt <= '0';
    fsm_rd_evt <= '0';
    fsm_addr_offs_rst <= '0';
    fsm_cmd_finish_evt <= '0';
    
    case fsm_state is 
      
      when idle =>
        if (cmd_outptr/=cmd_inptr and cmd_buffer(cmd_outptr).instr="000") then
          fsm_nextstate <= wr;
          fsm_addr_offs_rst <= '1';
        elsif (cmd_outptr/=cmd_inptr and cmd_buffer(cmd_outptr).instr="001") then
          fsm_nextstate <= rd;
          fsm_addr_offs_rst <= '1';
        end if;
        
      when wr =>
        fsm_wr_evt <= '1';
        if (mem_addr_offs=to_integer(unsigned(cmd_buffer(cmd_outptr).bl))) then
          fsm_nextstate <= idle;
          fsm_cmd_finish_evt <= '1'; 
        end if;
         
      when rd =>
        fsm_rd_evt <= '1';
        if (mem_addr_offs=to_integer(unsigned(cmd_buffer(cmd_outptr).bl))) then
          fsm_nextstate <= idle;
          fsm_cmd_finish_evt <= '1'; 
        end if;
        
    end case;
  end process;
  
  
  fsm_reg: process(px_mem_clk, px_cmd_rstn)
  begin
    if (px_cmd_rstn='0') then
      fsm_state <= idle;    
    elsif (px_mem_clk'event and px_mem_clk='1') then
      fsm_state <= fsm_nextstate;    
    end if;
  end process;
  
  
  fsm_outreg: process(px_mem_clk, px_cmd_rstn)
  begin
    if (px_cmd_rstn='0') then
      cmd_outptr    <= 0;
      wr_outptr     <= 0;
      rd_inptr      <= 0;
      mem_addr_offs <= 0;
      
    elsif (px_mem_clk'event and px_mem_clk='1') then
      
      if (fsm_addr_offs_rst='1') then
        mem_addr_offs <= 0;
        
      elsif (fsm_wr_evt='1') then
        mem_data(mem_addr) <= wr_buffer(wr_outptr);
        mem_addr_offs <= mem_addr_offs+1;
        if (wr_outptr=cMCB_WRRD_BUF_SIZE-1) then 
          wr_outptr <= 0;
        else
          wr_outptr <= wr_outptr+1;
        end if;
        
      elsif (fsm_rd_evt='1') then
        rd_buffer(rd_inptr) <= mem_data(mem_addr);
        mem_addr_offs <= mem_addr_offs+1;
        if (rd_inptr=cMCB_WRRD_BUF_SIZE-1) then
          rd_inptr <= 0;
        else
          rd_inptr <= rd_inptr+1;
        end if;
      end if;
      
      if (fsm_cmd_finish_evt='1') then
        if (cmd_outptr=cMCB_CMD_BUF_SIZE-1) then
          cmd_outptr <= 0;
        else
          cmd_outptr <= cmd_outptr+1;
        end if;
      end if;
            
    end if;
  end process;
  
  mem_addr_start <= to_integer(unsigned(cmd_buffer(cmd_outptr).addr));
  mem_addr       <= mem_addr_start + mem_addr_offs;
    

  cmd_buf: process(px_cmd_clk_i, px_cmd_rstn)
  begin
    if (px_cmd_rstn='0') then
      cmd_inptr <= 0;
      px_cmd_full_o <= '0';
      
    elsif (px_cmd_clk_i'event and px_cmd_clk_i='1') then
      if (px_cmd_en_i='1') then
        cmd_buffer(cmd_inptr).addr  <= px_cmd_addr_i(29 downto 2);
        cmd_buffer(cmd_inptr).bl    <= px_cmd_bl_i;
        cmd_buffer(cmd_inptr).instr <= px_cmd_instr_i;
        if (cmd_inptr=cMCB_CMD_BUF_SIZE-1) then
          cmd_inptr <= 0;
        else
          cmd_inptr <=  cmd_inptr+1;
        end if;
      end if;      
      
      if (cmd_outptr=(cmd_inptr+1) or (cmd_outptr=0 and cmd_inptr=cMCB_CMD_BUF_SIZE-1)) then
        if (px_cmd_en_i='1') then
          px_cmd_full_o <= '1';
        end if;
      elsif (cmd_outptr/=cmd_inptr) then
        px_cmd_full_o <= '0';
      end if;
            
    end if;  
  end process;
  

  wr_buf: process(px_wr_clk_i, px_cmd_rstn)
  begin
    if (px_cmd_rstn='0') then
      wr_inptr     <= 0;
      px_wr_full_o <= '0';
      
    elsif (px_wr_clk_i'event and px_wr_clk_i='1') then
      if (px_wr_en_i='1') then
        wr_buffer(wr_inptr) <= px_wr_data_i;
        if (wr_inptr=cMCB_WRRD_BUF_SIZE-1) then
          wr_inptr <= 0;
        else
          wr_inptr <= wr_inptr+1;
        end if;
      end if;      
      
      if (wr_outptr=(wr_inptr+1) or (wr_outptr=0 and wr_inptr=cMCB_WRRD_BUF_SIZE-1)) then
        if (px_wr_en_i='1') then
          px_wr_full_o <= '1';
        end if;
      elsif (wr_outptr/=wr_inptr) then
        px_wr_full_o <= '0';
      end if;        
            
    end if;  
  end process;
  
  
  rd_buf: process(px_rd_clk_i, px_cmd_rstn)
  begin
    if (px_cmd_rstn='0') then
      rd_outptr     <= 0;
      px_rd_empty_o <= '1';
      
    elsif (px_rd_clk_i'event and px_rd_clk_i='1') then
      if (px_rd_en_i='1') then
        if (rd_outptr=cMCB_WRRD_BUF_SIZE-1) then
          rd_outptr <= 0;
        else
          rd_outptr <=  rd_outptr+1;
        end if;
      end if;
      
      if ((rd_outptr+1)=rd_inptr or (rd_outptr=cMCB_WRRD_BUF_SIZE-1 and rd_inptr=0)) then
        if (px_rd_en_i='1') then
          px_rd_empty_o <= '1';
        end if;
      elsif (rd_outptr/=rd_inptr) then
        px_rd_empty_o <= '0';
      end if;
            
    end if;  
  end process;
  px_rd_data_o <= rd_buffer(rd_outptr);

end mcb_stim_a;